# ~\~ language=Python filename=parareal/__init__.py
# ~\~ begin <<lit/01-parareal.md|parareal/__init__.py>>[init]
from .tabulate_solution import tabulate
from .parareal import parareal
from . import abstract

__all__ = ["tabulate", "parareal", "schedule", "abstract"]
# ~\~ end



# ~\~ language=Python filename=parareal/tabulate_solution.py
# ~\~ begin <<lit/01-parareal.md|parareal/tabulate_solution.py>>[init]
from .abstract import (Solution, Vector)
from typing import (Sequence, Any)
import numpy as np

Array = Any

def tabulate(step: Solution, y_0: Vector, t: Array) -> Sequence[Vector]:
    """Tabulate the step-wise solution, starting from `y_0`, for every time
    point given in array `t`."""
    if isinstance(y_0, np.ndarray):
        return tabulate_np(step, y_0, t)

    y = [y_0]
    for i in range(1, t.size):
        y_i = step(y[i-1], t[i-1], t[i])
        y.append(y_i)
    return y

# ~\~ begin <<lit/01-parareal.md|tabulate-np>>[init]
def tabulate_np(step: Solution, y_0: Array, t: Array) -> Array:
    y = np.zeros(dtype=y_0.dtype, shape=(t.size,) + y_0.shape)
    y[0] = y_0
    for i in range(1, t.size):
        y[i] = step(y[i-1], t[i-1], t[i])
    return y
# ~\~ end
# ~\~ end



# ~\~ language=Python filename=parareal/harmonic_oscillator.py
# ~\~ begin <<lit/01-parareal.md|parareal/harmonic_oscillator.py>>[init]
from .abstract import (Problem)
from typing import Callable
from numpy.typing import NDArray
import numpy as np

def harmonic_oscillator(omega_0: float, zeta: float) -> Problem:
    def f(y, t):
        return np.r_[y[1], -2 * zeta * omega_0 * y[1] - omega_0**2 * y[0]]
    return f

# ~\~ begin <<lit/01-parareal.md|harmonic-oscillator-solution>>[init]
def underdamped_solution(omega_0: float, zeta: float) \
        -> Callable[[NDArray[np.float64]], NDArray[np.float64]]:
    amp   = 1 / np.sqrt(1 - zeta**2)
    phase = np.arcsin(zeta)
    freq  = omega_0 * np.sqrt(1 - zeta**2)

    def f(t: NDArray[np.float64]) -> NDArray[np.float64]:
        dampening = np.exp(-omega_0*zeta*t)
        q = amp * dampening * np.cos(freq * t - phase)
        p = - amp * omega_0 * dampening * np.sin(freq * t)
        return np.c_[q, p]
    return f
# ~\~ end

if __name__ == "__main__":
    import numpy as np  # type: ignore
    import pandas as pd  # type: ignore
    from plotnine import ggplot, geom_line, aes  # type: ignore

    from pintFoam.parareal.harmonic_oscillator import harmonic_oscillator
    from pintFoam.parareal.forward_euler import forward_euler
    from pintFoam.parareal.iterate_solution import iterate_solution
    from pintFoam.parareal.tabulate_solution import tabulate_np

    OMEGA0 = 1.0
    ZETA = 0.5
    H = 0.001
    system = harmonic_oscillator(OMEGA0, ZETA)

    def coarse(y, t0, t1):
        return forward_euler(system)(y, t0, t1)

    # fine :: Solution[NDArray]
    def fine(y, t0, t1):
        return iterate_solution(forward_euler(system), H)(y, t0, t1)

    y0 = np.array([1.0, 0.0])
    t = np.linspace(0.0, 15.0, 100)
    exact_result = underdamped_solution(OMEGA0, ZETA)(t)
    euler_result = tabulate_np(fine, y0, t)

    data = pd.DataFrame({
        "time": t,
        "exact_q": exact_result[:,0],
        "exact_p": exact_result[:,1],
        "euler_q": euler_result[:,0],
        "euler_p": euler_result[:,1]})

    plot = ggplot(data) \
        + geom_line(aes("time", "exact_q")) \
        + geom_line(aes("time", "euler_q"), color="#000088")
    plot.save("plot.svg")

# ~\~ end



# ~\~ language=Python filename=parareal/futures.py
# ~\~ begin <<lit/02-parafutures.md|parareal-futures>>[init]
from .abstract import (Solution, Mapping, Vector)
from typing import (Callable)
from dataclasses import dataclass
from math import ceil
import numpy as np
from numpy.typing import NDArray
from dask.distributed import Client, Future  # type: ignore
import logging


def identity(x):
    return x

def pairs(lst):
    return zip(lst[:-1], lst[1:])
# ~\~ end
# ~\~ begin <<lit/02-parafutures.md|parareal-futures>>[1]
def combine(c1: Vector, f1: Vector, c2: Vector) -> Vector:
    return c1 + f1 - c2
# ~\~ end
# ~\~ begin <<lit/02-parafutures.md|parareal-futures>>[2]
@dataclass
class Parareal:
    client: Client
    coarse: Callable[[int], Solution]
    fine: Callable[[int], Solution]
    c2f: Mapping = identity
    f2c: Mapping = identity

    def _c2f(self, x: Future) -> Future:
        if self.c2f is identity:
            return x
        return self.client.submit(self.c2f, x)

    def _f2c(self, x: Future) -> Future:
        if self.f2c is identity:
            return x
        return self.client.submit(self.f2c, x)

    def _coarse(self, n_iter: int, y: Future, t0: float, t1: float) ->  Future:
        logging.debug("Coarse run: %s, %s, %s", y, t0, t1)
        return self.client.submit(self.coarse(n_iter), y, t0, t1)

    def _fine(self, n_iter: int, y: Future, t0: float, t1: float) -> Future:
        logging.debug("Fine run: %s, %s, %s", y, t0, t1)
        return self.client.submit(self.fine(n_iter), y, t0, t1)

    # ~\~ begin <<lit/02-parafutures.md|parareal-methods>>[init]
    def step(self, n_iter: int, y_prev: list[Future], t: NDArray[np.float64]) -> list[Future]:
        m = t.size
        y_next = [None] * m
        y_next[0] = y_prev[0]

        for i in range(1, m):
            c1 = self._c2f(self._coarse(n_iter, self.f2c(y_next[i-1]), t[i-1], t[i]))
            f1 = self._fine(n_iter, y_prev[i-1], t[i-1], t[i])
            c2 = self._c2f(self._coarse(n_iter, self.f2c(y_prev[i-1]), t[i-1], t[i]))
            y_next[i] = self.client.submit(combine, c1, f1, c2)

        return y_next
    # ~\~ end
    # ~\~ begin <<lit/02-parafutures.md|parareal-methods>>[1]
    def schedule(self, y_0: Vector, t: NDArray[np.float64]) -> list[list[Future]]:
        # schedule initial coarse integration
        y_init = [self.client.scatter(y_0)]
        for (a, b) in pairs(t):
            y_init.append(self._coarse(0, y_init[-1], a, b))

        # schedule all iterations of parareal
        jobs = [y_init]
        for n_iter in range(len(t)):
            jobs.append(self.step(n_iter+1, jobs[-1], t))

        return jobs
    # ~\~ end
    # ~\~ begin <<lit/02-parafutures.md|parareal-methods>>[2]
    def wait(self, jobs, convergence_test):
        for i in range(len(jobs)):
            result = self.client.gather(jobs[i])
            if convergence_test(result):
                for j in jobs[i+1:]:
                    self.client.cancel(j, force=True)
                return result
        return result
    # ~\~ end
# ~\~ end



# ~\~ language=Python filename=parareal/parareal.py
# ~\~ begin <<lit/01-parareal.md|parareal/parareal.py>>[init]
from .abstract import (Solution, Mapping)
import numpy as np

def identity(x):
    return x

def parareal(
        coarse: Solution,
        fine: Solution,
        c2f: Mapping = identity,
        f2c: Mapping = identity):
    def f(y, t):
        m = t.size
        y_n = [None] * m
        y_n[0] = y[0]
        for i in range(1, m):
            # ~\~ begin <<lit/01-parareal.md|parareal-core-2>>[init]
            y_n[i] = c2f(coarse(f2c(y_n[i-1]), t[i-1], t[i])) \
                   + fine(y[i-1], t[i-1], t[i]) \
                   - c2f(coarse(f2c(y[i-1]), t[i-1], t[i]))
            # ~\~ end
        return y_n
    return f

def parareal_np(
        coarse: Solution,
        fine: Solution,
        c2f: Mapping = identity,
        f2c: Mapping = identity):
    def f(y, t):
        m = t.size
        y_n = np.zeros_like(y)
        y_n[0] = y[0]
        for i in range(1, m):
            # ~\~ begin <<lit/01-parareal.md|parareal-core-2>>[init]
            y_n[i] = c2f(coarse(f2c(y_n[i-1]), t[i-1], t[i])) \
                   + fine(y[i-1], t[i-1], t[i]) \
                   - c2f(coarse(f2c(y[i-1]), t[i-1], t[i]))
            # ~\~ end
        return y_n
    return f
# ~\~ end



# ~\~ language=Python filename=parareal/iterate_solution.py
# ~\~ begin <<lit/01-parareal.md|parareal/iterate_solution.py>>[init]
from .abstract import (Vector, Solution)
import numpy as np
import math

def iterate_solution(step: Solution, h: float) -> Solution:
    def iter_step(y: Vector, t_0: float, t_1: float) -> Vector:
        """Stepping function of iterated solution."""
        n = math.ceil((t_1 - t_0) / h)
        steps = np.linspace(t_0, t_1, n + 1)
        for t_a, t_b in zip(steps[:-1], steps[1:]):
            y = step(y, t_a, t_b)
        return y
    return iter_step
# ~\~ end



# ~\~ language=Python filename=parareal/abstract.py
# ~\~ begin <<lit/01-parareal.md|parareal/abstract.py>>[init]
from __future__ import annotations
from typing import (Callable, Protocol, TypeVar, Union)

# ~\~ begin <<lit/01-parareal.md|abstract-types>>[init]
TVector = TypeVar("TVector", bound="Vector")

class Vector(Protocol):
    def __add__(self: TVector, other: TVector) -> TVector:
        ...

    def __sub__(self: TVector, other: TVector) -> TVector:
        ...

    def __mul__(self: TVector, other: float) -> TVector:
        ...

    def __rmul__(self: TVector, other: float) -> TVector:
        ...

# ~\~ end
# ~\~ begin <<lit/01-parareal.md|abstract-types>>[1]
Mapping = Callable[[TVector], TVector]
# ~\~ end
# ~\~ begin <<lit/01-parareal.md|abstract-types>>[2]
Problem = Callable[[TVector, float], TVector]
# ~\~ end
# ~\~ begin <<lit/01-parareal.md|abstract-types>>[3]
Solution = Union[Callable[[TVector, float, float], TVector],
                 Callable[..., TVector]]
# ~\~ end
# ~\~ end



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['parareal']

package_data = \
{'': ['*']}

install_requires = \
['argh>=0.26.2,<0.27.0',
 'dask-mpi>=2022.4.0,<2023.0.0',
 'dask[complete]>=2022.10.0,<2023.0.0',
 'h5py>=3.7.0,<4.0.0',
 'mpi4py>=3.1.3,<4.0.0',
 'numpy>=1.23.4,<2.0.0']

setup_kwargs = {
    'name': 'parareal',
    'version': '0.1.0',
    'description': 'Parallel-in-time framework for Python',
    'long_description': '---\ntitle: Parareal\nsubtitle: a library for parallel-in-time computations in Python\n---\n[![Entangled badge](https://img.shields.io/badge/entangled-Use%20the%20source!-%2300aeff)](https://entangled.github.io/)\n\nThis package runs the parareal algorithm on a black-box simulator. Parallelisation is managed through Dask.\n\n# License\nApache 2, see `LICENSE`.\n\n',
    'author': 'Johan Hidding',
    'author_email': 'j.hidding@esciencecenter.nl',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<3.11',
}


setup(**setup_kwargs)



# ~\~ language=Python filename=parareal/forward_euler.py
# ~\~ begin <<lit/01-parareal.md|parareal/forward_euler.py>>[init]
from .abstract import (Vector, Problem, Solution)

def forward_euler(f: Problem) -> Solution:
    """Forward-Euler solver."""
    def step(y: Vector, t_0: float, t_1: float) -> Vector:
        """Stepping function of Euler method."""
        return y + (t_1 - t_0) * f(y, t_0)
    return step
# ~\~ end
